/*#======================================================================
# This source code is public domain, and may be freely distributed,
# modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
# WARRANTY OF ANY KIND.
#======================================================================
# _DY_CONTRIB_PGM_:du_smol2smi:Convert MOL-file to SMILES or TDTs
#======================================================================
# LICENSE MODULES: toolkits:
#                  programs:
#                  servers:
# DESCRIPTION: 
# Convert MOL-file to SMILES or TDTs
#
# This program converts MDL format molfiles to Daylight SMILES and
# TDTs.  SD (structure-data) files are converted to TDTS with SMILES
# and other data.
#
# Stereochemistry is inferred from the molfile and output as an 
# Isomeric SMILES.  Atom "parity", "up/down bonds", 2D and 3D
# coordinates are all taken into account in accordance with the
# program options specified.
#=====================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include "dt_smiles.h"
#include "du_utils.h"
#include "du_mol2smi.h"
#include "du_3d.h"

#define MAX(x,y)  (((x) > (y)) ?  (x) : (y))
#define MIN(x,y)  (((x) < (y)) ?  (x) : (y))
#define XOR(a,b)  ((a) && !(b) || !(a) && (b))

#define MAXLINE  220   /*** Max length of molfile line (?) ***/

#define INT 1       /*** For use with strn2num() ***/
#define FLOAT 2     /*** For use with strn2num() ***/
#define STRING 3    /*** For use with strn2num() ***/

/************************************************************
 *** Unifying struct for atom-tuple info
 ************************************************************/
typedef struct {
  dt_Handle h;
  float     x;                 /*** x coord (2/3D)    ***/
  float     y;                 /*** y coord (2/3D)    ***/
  float     z;                 /*** z coord (3D)      ***/
  int       parity;            /*** stereo parity     ***/
  int       hhh;               /*** (hcount + 1)      ***/
  int       visited;           /*** internal flag     ***/
  char      symbol[4];         /*** MDL atom symbol   ***/
  int       valence;
  int       mapclass;
  int       radcode;
} ATOMINFO;

/************************************************************
 *** Unifying struct for bond-tuple info
 ************************************************************/
typedef struct {
  dt_Handle h;
  dt_Handle from_atom;
  dt_Handle to_atom;
  int stereo;
} BONDINFO;

/********************************************************************
***  Function Declarations:
***     symb2numb -- atomic symbol -> atomic number
***    atnum2mass -- atomic number -> atomic mass
***    stream2seq -- copy stream to seq so we can mod mol
*** signed_volume -- chirality determinant
***    setchiral1 -- chirality from MDL parity
***    setchiral2 -- chirality from 3D or (2D + up/down)
***   setcistrans -- cis/trans from 2D/3D
***      strn2num -- convert n chars of string to number
***      rd_normh -- return the normal hydrogen count of an atom
***    rd_valence -- return the valence for an atom
*** radical_rings -- correct molecule for radical labels
********************************************************************/

static int   symb2numb     (char *symb);
static int   atnum2mass    (int atnum);
dt_Handle stream2seq(dt_Handle stream);
static float signed_volume (dt_Handle atom, dt_Handle bondstream);
static int   setchiral1    (dt_Handle mol);
static int   setchiral2    (dt_Handle mol);
static int   setcistrans   (dt_Handle mol, int ring_cistrans);
static int   strn2num      (char *str, int len, int type, void *num);
static int   rd_normh      (dt_Handle atom, int fixval, int radcode);
static int   rd_valence    (dt_Handle atom, int vbond);
static int   radical_rings (dt_Handle mol);

/************************************************************
*** Global variables.
************************************************************/

ATOMINFO *atoms = NULL;       /*** Atom info array (dynamic)      ***/
BONDINFO *bonds = NULL;       /*** Bond info array (dynamic)      ***/
int natoms;                   /*** Number of atoms (from molfile) ***/
int nbonds;                   /*** Number of bonds (from molfile) ***/
int nnewhs;                   /*** Number of new explicit H's     ***/

int max_natoms = 0;         /*** Max number of atoms so far     ***/
int max_nbonds = 0;         /*** Max number of bonds so far     ***/

/*======================================================================
 * User-settable global properties.
 ====================================================================== */

int parity_over_bonds = 0;  /*** Boolean: Stereo precedence     ***/
int ring_cistrans = 0;      /*** Boolean: cis-trans ringbonds?  ***/
int m__iso_are_weights = 0; /*** Boolean: weights not defects?  ***/

void du_mol2smi_setparity(int val)
{
  parity_over_bonds = val;
  return;
}

void du_mol2smi_setringEZ(int val)
{
  ring_cistrans = val;
  return;
}

void du_mol2smi_setisoweight(int val)
{
  m__iso_are_weights = val;
  return;
}

/***************************************************************************
* FUNCTION: du_mol2smi
*
* DESCRIPTION:
*	Convert an SD record to a molecule object.
*
* RETURNS: (dt_Handle)
***************************************************************************/

dt_Handle du_mol2smi(FILE *inputfl, int *rc)
{
  dt_Handle atom, atomstream, mol;
  char errortext[100] = "";
  char symb[4] = "";
  
  char *line, *p;
  
  int at_num, bond_type, bond_from, bond_to, bond_stereo;
  int charge, defect, hcount, i, j, k, len, weight, radcode;
  int got2d, got3d, gotsymbol;

  nbonds = natoms = nnewhs = 0;
  
  /***********************************************
  ***  Counts line of MOL-file:
  ***  aaabbblllfffcccsssxxxrrrpppiiimmmvvvvvv
  ***  1234567890123456789012345678901234567890
  ***           |10       |20       |30       |40
  ***********************************************/
  if (!(line = du_fgetline(&len, inputfl)))
    {
      fprintf(stderr, "Unexpected EOF...\n");
      *rc = -1;
      return (NULL_OB);
    }
  if (!strn2num(line, 3, INT, &natoms))
    {
      fprintf(stderr, "ERROR (MOL2SMI) Can't read natoms: %s\n ", line);
      *rc = -1;
      return (NULL_OB);
    }
  if (!strn2num(line + 3, 3, INT, &nbonds))
    {
      fprintf(stderr, "ERROR (MOL2SMI) Can't read nbonds: %s\n", line);
      *rc = -1;
      return (NULL_OB);
    }
    
  /***********************************************************
  *** Handle special case of natoms = 0.
  *** Non-struct IFF *rc==0 AND mol==NULL_OB.
  ***********************************************************/
  if (0 == natoms)
    {
      /* fprintf(stderr, 
	 "NOTE (MOL2SMI): Non-struct (natoms = 0): %s\n", line); */
      *rc = 0;
      return (NULL_OB);
    }

  /***********************************************************
  *** (Re)allocate ATOMINFO array of structs
  *** Allocate at least 32 atoms.  Allocate 3*natoms to
  *** allow for added hydrogens (kludge).
  ***********************************************************/
  
  if (!atoms)
    {
      max_natoms = MAX(32, 3 * natoms);
      atoms = (ATOMINFO *)malloc((max_natoms + 1) * sizeof(ATOMINFO));
    }
  else if (3 * natoms > max_natoms)
    {
      max_natoms = 3 * natoms;
      atoms = (ATOMINFO *)realloc(atoms, (max_natoms + 1) * sizeof(ATOMINFO));
    }

  /*** Always zero this array, whether new or not. ***/
  memset(atoms, '\0', (max_natoms + 1) *  sizeof(ATOMINFO));
  
  /***********************************************************
  *** (Re)allocate BONDINFO array of structs
  *** Allocate at least 32 bonds.  Allocate 3*nbonds to
  *** allow for added hydrogens (kludge).
  ***********************************************************/
  
  if (!bonds)
    {
      max_nbonds = MAX(32, 3 * nbonds);
      bonds = (BONDINFO *)malloc((max_nbonds + 1) * sizeof(BONDINFO));
    }
  else if (3 * nbonds > max_nbonds)
    {
      max_nbonds = 3 * nbonds;
      bonds = (BONDINFO *)realloc(bonds, (max_nbonds + 1) *
				  sizeof(BONDINFO));
    }

  /***********************************************************
  *** Allocate and mod_on molecule.
  ***********************************************************/
  
  mol = dt_alloc_mol();
  dt_mod_on(mol);
  
  /********************************************************************
  ***  ATOM LOOP: Note that atom indices start at one...  If an
  ***  atom is not interpreted, go to next CT.  Charge is decoded
  ***  by the formula: charge = (4 - charge_code)  Note that
  ***  atoms are added with (hcount) implicit hydrogens.
  ***
  ***  The format is:
  ***  xxxxx.xxxxyyyyy.yyyyzzzzz.zzzz aaaddcccssshhhbbbvvv
  ***  12345678901234567890123456789012345678901234567890123456789
  ***           |10       |20       |30       |40       |50        
  ***
  ***  xyz's   - coordinates           sss    - stereo parity 
  ***  aaa     - atom symbol           hhh    - (hcount + 1)
  ***  dd      - mass difference       bbb    - stereo care box
  ***  ccc     - (4 - charge)          vvv    - valence
  ********************************************************************/
  
  for (i = 1; i <= natoms; i++)
    {
      if (!(line = du_fgetline(&len, inputfl)))
	{
	  strcpy(errortext, "Unexpected EOF...");
	  *rc = -1;
	  return (NULL_OB);
	}
      *(symb + 3) = '\0';

      /*** JJD - Handle short lines ***/
      if ((len < 34) ||
	  ! strn2num(line, 10, FLOAT, &atoms[i].x)         ||
	  ! strn2num(line + 10, 10, FLOAT, &atoms[i].y)    ||
	  ! strn2num(line + 20, 10, FLOAT, &atoms[i].z)    ||
	  ! strn2num(line + 31, 3, STRING, symb)           ||
	  ! strn2num(line + 34, 2, INT, &defect)           ||
	  ! strn2num(line + 36, 3, INT, &charge)           ||
	  ! strn2num(line + 39, 3, INT, &atoms[i].parity)  ||
	  ! strn2num(line + 42, 3, INT, &atoms[i].hhh)     ||
	  ! strn2num(line + 48, 3, INT, &atoms[i].valence) ||
	  ! strn2num(line + 60, 3, INT, &atoms[i].mapclass))
	{
	  fprintf(stderr, "ERROR (MOL2SMI) Faulty atom record: %s\n", line);
	  dt_dealloc(mol);
	  *rc = -1;
	  return (NULL_OB);
	}
      
      /*** Handle short lines ***/

      if( len < 36 ) defect = 0;
      if( len < 39 ) charge = 0;
      if( len < 42 ) atoms[i].parity = 0;
      if( len < 45 ) atoms[i].hhh = 0;
      if( len < 51 ) atoms[i].valence = 0;
      if( len < 63 ) atoms[i].mapclass = 0;

      /*** Deal with the symbol.  Either get the proper atomic number
	for the symbol, or save the symbol string and substitute with
	the wildcard atom (*). ***/

      if (p = strchr(symb, ' ')) *p = '\0';
      if      (!strcmp("D", symb)) { strcpy(symb, "H"); weight = 2; }
      else if (!strcmp("T", symb)) { strcpy(symb, "H"); weight = 3; }
      else                                            { weight = 0; }
      at_num = symb2numb(symb);
      if (!at_num)
	{
	  fprintf(stderr,
		  "NOTE: \"%s\" atom translated to \"*\"\n", symb);
	  strcpy(atoms[i].symbol, symb);
	  dt_setboolean(mol, 9, "gotsymbol", TRUE);
	}

      /*** Check for coordinates, to use for stereo perception ***/

      if (fabs(atoms[i].z) > FLT_EPSILON)
	dt_setboolean(mol, 5, "got3d", TRUE);
      if ((fabs(atoms[i].x) > FLT_EPSILON) ||
	  (fabs(atoms[i].y) > FLT_EPSILON))
	dt_setboolean(mol, 5, "got2d", TRUE);
      
      if (atoms[i].hhh) 
	hcount = atoms[i].hhh - 1;
      else 
	hcount = 0;
	
      atoms[i].h = dt_addatom(mol, at_num, hcount);
      dt_setinteger(atoms[i].h, 5, "index", i);
      
      if (defect)
	weight = atnum2mass(at_num) + defect;
      if (weight)
	dt_setweight(atoms[i].h, weight);

      if (charge == 4)
	atoms[i].radcode = 1;	/* Add radical post-processing */

      /*** JJD - 6/2000  +5 charge is represented as '-1'
	   (not documented in MDL docs) ***/
      if (charge != 0)
	charge = (4 - charge);
      if (charge)
	dt_setcharge(atoms[i].h, charge);
    }
  
  /********************************************************************
  ***  BOND LOOP: 
  ***  
  ***  The format is:
  ***  111222tttsssxxxrrrccc
  ***  12345678901234567890123456789
  ***           |10       |20       
  ********************************************************************/
  
  for (i = 1; i <= nbonds; i++)
    {
      if (!(line = du_fgetline(&len, inputfl)))
	{
	  strcpy(errortext, "Unexpected EOF...");
	  dt_dealloc(mol);
	  *rc = -1;
	  return (NULL_OB);
	}
      /*** JJD - Handle short lines ***/
      if ((len < 9) ||
	  ! strn2num(line, 3, INT, &bond_from)     ||
	  ! strn2num(line + 3, 3, INT, &bond_to)   ||
	  ! strn2num(line + 6, 3, INT, &bond_type) ||
	  ! strn2num(line + 9, 3, INT, &bond_stereo))
	{
	  fprintf(stderr, "ERROR (MOL2SMI) Faulty bond record: %s\n", line);
	  dt_dealloc(mol);
	  *rc = -1;
	  return (NULL_OB);
	}
      
      /*** Handle short lines ***/
      if( len < 12 ) bond_stereo = 0;

      /*********************************************
      *** We shouldn't allow bond_type > 3, but
      *** this gives some non-standard molfiles
      *** a chance.
      *********************************************/

      if (bond_type > 4) bond_type = 4;
      bonds[i].h = dt_addbond(atoms[bond_from].h, 
			      atoms[bond_to].h, bond_type);
      bonds[i].from_atom = atoms[bond_from].h;
      bonds[i].to_atom = atoms[bond_to].h;
      bonds[i].stereo = bond_stereo;
      dt_setinteger(bonds[i].h, 5, "index", i);
    }
  
  /********************************************************************
   ***  Check for "M  CHG", "M  ISO", "M  RAD", or "M  END".  These
   ***  are processed here.  Anything else ("$$$$", "> ...", "$MOL"
   ***  terminates the processing.
   ***
   ***  In some cases, SD records in RD files are improperly
   ***  terminated (no "M  END" or "$$$$").  Hence, we have to be
   ***  clever and peek ahead at the next line, lest we read past
   ***  the end of the SD record.
   ********************************************************************/

  while (1)
    {
      /*** Peek at the first character of the next line ***/

      line[0] = getc(inputfl);
      ungetc(line[0], inputfl);

      /*** If the next line doesn't begin with 'M' we're done ***/

      if (line[0] != 'M')
	break;

      /*** Process the next data. ***/

      line = du_fgetline(&len, inputfl);
      if (line == NULL) break;

      if (!strncmp(line, "M  END", 6))
	break;

      if (!strncmp(line, "M  CHG", 6))
	{
	  p = line + 6;
	  sscanf(p, "%d", &j);
	  p += 3;
	  for (k = 1; k <= j; ++k)
	    {
	      sscanf(p, "%d%d", &i, &charge);
	      if (i > natoms)
		{
		  fprintf(stderr,
			  "WARNING (MOL2SMI) Bad CHG line: %s\n", line);
		  break;
		}
	      dt_setcharge(atoms[i].h, charge);
	      p += 8;
	    }
	}
      else if (!strncmp(line, "M  ISO", 6))
	{
	  p = line + 6;
	  sscanf(p, "%d", &j);
	  p += 3;
	  for (k=1; k <= j; ++k) {
	    sscanf(p, "%d%d", &i, &defect);
	    if (i > natoms)
	      {
		fprintf(stderr,
			"WARNING (MOL2SMI) Bad ISO line: %s\n", line);
		break;
	      }
	    
	    /***********************************************************
	    ***  defect is supposed to be a delta according to MDL docs.
	    ***  In practice in MDL db's it is the weight itself.
	    ***********************************************************/
	    
	    if (m__iso_are_weights)
	      weight = defect;
	    else
	      weight = atnum2mass(dt_number(atoms[i].h)) + defect;
	    dt_setweight(atoms[i].h, weight);
	    
	    p += 8;
	  }
	}
      else if (!strncmp(line, "M  RAD", 6))
	{
	  p = line + 6;
	  sscanf(p, "%d", &j);
	  p += 3;
	  for (k=1; k <= j; ++k)
	    {
	      sscanf(p, "%d%d", &i, &radcode);
	      if (i > natoms)
		{
		  fprintf(stderr, "ERROR (MOL2SMI) i > natoms: \"%s\";\n",
			  line);
		  break;
		}
	      atoms[i].radcode = radcode;
	      p += 8;
	    }
	}
      else
	fprintf(stderr, "WARNING (MOL2SMI) Uninterpreted line: \"%s\"; ",
		line);
    }
  
  /***************************************************
  ***  H count adjustments - based on MDL valences.
  ***************************************************/
    
  for (i = 1; i <= natoms; i++)
    {
      if (atoms[i].valence == 0)
	dt_setimp_hcount(atoms[i].h, rd_normh(atoms[i].h, -1, 
					      atoms[i].radcode));
      else if (15 > atoms[i].valence)
	dt_setimp_hcount(atoms[i].h, rd_normh(atoms[i].h, atoms[i].valence,
					      atoms[i].radcode));
      else
	dt_setimp_hcount(atoms[i].h, 0);
      
      if (atoms[i].hhh)
	dt_setimp_hcount(atoms[i].h, atoms[i].hhh - 1);
    }
  
  /******************************************************************
  *** Stereo stuff.   
  *** setchiral1 infers tetrahedral chirality from MDL atom parity.
  *** setchiral2 infers tetrahedral chirality from MDL 2D + up/down
  *** bonds or 3D coordinates.
  *** If setchiral1 is called first it takes precedence over
  *** setchiral2 ("parity over bonds").
  *** setcistrans infers cis/trans chirality from 2D or 3D coordinates.
  ******************************************************************/
  
  if (parity_over_bonds) setchiral1(mol);
  if (dt_boolean(mol, 5, "got2d") || dt_boolean(mol, 5, "got3d"))
    {
      setchiral2(mol);
      setcistrans(mol, ring_cistrans);
    }
  if (!parity_over_bonds) setchiral1(mol);
  
  /****************************************************
  ***  Make SMILES
  ****************************************************/

  radical_rings(mol);
  
  if (!dt_mod_off(mol))
    {
      fprintf(stderr, "ERROR (MOL2SMI) dt_mod_off failure.\n");
      *rc = -1;
      return (NULL_OB);
    }


  got2d = dt_boolean(mol, 5, "got2d");
  got3d = dt_boolean(mol, 5, "got3d");
  gotsymbol = dt_boolean(mol, 9, "gotsymbol");

  atomstream = dt_stream(mol, TYP_ATOM);
  while (NULL_OB != (atom = dt_next(atomstream)))
    {
      i = dt_integer(atom, 5, "index");
      dt_setinteger(atom, 3, "map", atoms[i].mapclass);

      if (got2d)
	{
	  dt_setreal(atom, 1, "x", atoms[i].x);
	  dt_setreal(atom, 1, "y", atoms[i].y);
	  if (got3d)
	    dt_setreal(atom, 1, "z", atoms[i].z);
	}

      if (gotsymbol)
	dt_setstring(atom, 6, "symbol", strlen(atoms[i].symbol), 
		     atoms[i].symbol);

      dt_setinteger(atom, 5, "index", 0); /* Clear the index */
    }
  dt_dealloc(atomstream);
  *rc = 0;
  return (mol);
} /*** End of molecule loop ***/




/********************************************************************
*** du_mol2d3d - return string with xyz's for D3D datatype
********************************************************************/
char *du_mol2d3d(dt_Handle mol)
{
  dt_Handle atomstream, atom, bondstream, bond, atom2;
  int acount, hcount_total, bufsize, bytes;
  char *p;
  static int oldbufsize = 0;
  static char *buf3d = NULL;

  hcount_total = 0;
  atomstream = dt_canatom_stream(mol, 0, 0);
  acount = dt_count(atomstream, TYP_ATOM);
  while (NULL_OB != (atom = dt_next(atomstream)))
    {
      hcount_total += dt_hcount(atom);
      dt_setboolean(atom, 7, "visited", FALSE);
    }
  
  if (natoms == acount + hcount_total)
    bufsize = 24 * natoms + 1;
  else
    bufsize = 24 * acount + 1;
  
  if (!buf3d)
    buf3d = (char *) malloc(bufsize);
  else if (bufsize > oldbufsize)
    buf3d = (char *) realloc(buf3d, bufsize);
  else
    bufsize = oldbufsize;
  oldbufsize = bufsize;

  dt_reset(atomstream);
  for (p = buf3d; NULL_OB != (atom = dt_next(atomstream)); p += bytes)
    {
      bytes = sprintf(p, "%.2f,%.2f,%.2f,", dt_real(atom, 1, "x"),
		      dt_real(atom, 1, "y"), dt_real(atom, 1, "z"));
      dt_setboolean(atom, 7, "visited", TRUE);
    }

  /*************************************************
  ***  Now for the H's -- if we have ALL coordinates
  *************************************************/
  if (natoms == acount + hcount_total)
    {
      dt_reset(atomstream);
      while (NULL_OB != (atom = dt_next(atomstream)))
	{
	  bondstream = dt_stream(atom, TYP_BOND);
	  while (NULL_OB != (bond = dt_next(bondstream)))
	    {
	      atom2 = dt_xatom(atom, bond);
	      if (dt_number(atom2) == DX_ATN_H)
		{
		  if (!dt_boolean(atom2, 7, "visited"))
		    {
		      bytes = sprintf(p, "%.2f,%.2f,%.2f,",
				      dt_real(atom2, 1, "x"),
				      dt_real(atom2, 1, "y"),
				      dt_real(atom2, 1, "z"));
		      dt_setboolean(atom2, 7, "visited", TRUE);
		      p += bytes;
		    }
		}
	    }
	  dt_dealloc(bondstream);
	}
    }
  *(--p) = '\0'; /*** Terminate at trailing comma ***/

  dt_dealloc(atomstream);
  return (buf3d);
}

/********************************************************************
*** du_mol2d2d - return string with xy's for 2D datatype
********************************************************************/
char *du_mol2d2d(dt_Handle mol)
{
  dt_Handle atomstream, atom;
  int bufsize, bytes;
  char *p;
  static char *buf2d = NULL;
  static int oldbufsize = 0;
  
  atomstream = dt_canatom_stream(mol, 0, 0);
  bufsize   = 16 * dt_count(atomstream, TYP_ATOM) + 1;
  if (!buf2d)
    buf2d = (char *) malloc(bufsize);
  else if (bufsize > oldbufsize) 
    buf2d = (char *) realloc(buf2d, bufsize);
  else 
    bufsize = oldbufsize;
  oldbufsize = bufsize;

  for (p = buf2d; (NULL_OB != (atom = dt_next(atomstream))); p += bytes)
    bytes = sprintf(p, "%.2f,%.2f,",
		    dt_real(atom, 1, "x"),
		    dt_real(atom, 1, "y"));

  *(--p) = '\0'; /*** Terminate at trailing comma ***/

  dt_dealloc(atomstream);
  return (buf2d);
}


/*======================================================================
 * Internal functions
 ====================================================================== */

/********************************************************************
 ***  symb2numb - Note that "*" is the "unknown atom".
 ********************************************************************/

static int symb2numb(char *symb)
{
  int i;
  char s[5], *p;
  static char *symbol_table[] = {
    "*",
    "H",  "He", "Li", "Be", "B",  "C",  "N",  "O",  "F",  "Ne", /* 1-10 */
    "Na", "Mg", "Al", "Si", "P",  "S",  "Cl", "Ar", "K",  "Ca", /* 11-20 */
    "Sc", "Ti", "V",  "Cr", "Mn", "Fe", "Co", "Ni", "Cu", "Zn", /* 21-30 */
    "Ga", "Ge", "As", "Se", "Br", "Kr", "Rb", "Sr", "Y",  "Zr", /* 31-40 */
    "Nb", "Mo", "Tc", "Ru", "Rh", "Pd", "Ag", "Cd", "In", "Sn", /* 41-50 */
    "Sb", "Te", "I",  "Xe", "Cs", "Ba", "La", "Ce", "Pr", "Nd", /* 51-60 */
    "Pm", "Sm", "Eu", "Gd", "Tb", "Dy", "Ho", "Er", "Tm", "Yb", /* 61-70 */
    "Lu", "Hf", "Ta", "W",  "Re", "Os", "Ir", "Pt", "Au", "Hg", /* 71-80 */
    "Tl", "Pb", "Bi", "Po", "At", "Rn", "Fr", "Ra", "Ac", "Th", /* 81-90 */
    "Pa", "U",  "Np", "Pu", "Am", "Cm", "Bk", "Cf", "Es", "Fm", /* 91-100 */
    "Md", "No", "Lr", "Rf", "Ha", "dummy" };                    /* 101-106 */

  strcpy(s, symb);
  if ((NULL != (p = strchr(s, '+'))) ||
      (NULL != (p = strchr(s, '-'))))
    *p = '\0';

  for (i = 0; (i < 106) && (strcmp(s, symbol_table[i]));  ++i)
    ;
  if (i > 105) i = 0;
  return(i);
}

/********************************************************************
***  atnum2mass - Return mass given atomic number.
***  Returns total number of neutron and protons for given atnum.
***  If naturally occurring, for the most common isotope.
***  If not naturally occurring, for the most stable isotope.
***  Zero (0) if not known - error.
***  Reference: CRC Handbook of Chemistry and Physics, 60th Ed.
********************************************************************/

static int atnum2mass(int atnum)
{
  static int mass[] = {
    0,
    1,   4,   7,   9,   11,  12,  14,  16,  19,  20,   /*  1-10 */
    23,  24,  27,  28,  31,  32,  35,  40,  39,  40,   /* 11-20 */
    45,  48,  51,  52,  55,  56,  59,  58,  63,  64,   /* 21-30 */
    69,  74,  75,  80,  79,  84,  85,  88,  89,  90,   /* 31-40 */
    93,  98,  97,  102, 103, 106, 107, 114, 115, 120,  /* 41-50 */
    121, 130, 127, 132, 133, 138, 139, 140, 141, 142,  /* 51-60 */
    145, 152, 153, 158, 159, 164, 165, 166, 169, 174,  /* 61-70 */
    175, 180, 181, 184, 187, 192, 193, 195, 197, 202,  /* 71-80 */
    205, 208, 209, 209, 210, 222, 223, 226, 227, 232,  /* 81-90 */
    231, 238, 237, 244, 243, 247, 247, 251, 254, 257,  /* 91-100 */
    257, 255, 256, 260, 262, 0 };                      /* 101-106 */

  if ((atnum < 1) || (atnum > 105)) {
    fprintf(stderr,
      "ERROR: Unknown atom number --> %d\n", atnum);
    return (0);
  }
  return (mass[atnum]);
}

/********************************************************************
***  stream2seq - Copy a stream to a new sequence.
********************************************************************/

dt_Handle stream2seq(dt_Handle stream)
{
  dt_Handle ob, seq = dt_alloc_seq();

  dt_reset(stream);
  while (NULL_OB != (ob = dt_next(stream))) dt_append(seq, ob);
  return seq;
}

/********************************************************************
***  signed_volume - compute volume defined by the xyzs
***  of the four atoms single-bonded to the chiral atom. Sign
***  of the volume determines R/S chirality.
***  (Some 2.5D implicit H's not handled.)
********************************************************************/

static float signed_volume(dt_Handle atom, dt_Handle bondseq)
{
  dt_Handle atom2;
  float   a[3], b[3], c[3], d[3];
  int     i, imp_hcount;

  switch (dt_count(bondseq, TYP_BOND)) {
    case 3: imp_hcount = 1; break;
    case 4: imp_hcount = 0; break;
    default: return (0.0);
  }

  /**********************************************************
  *** Extract xyz's of attached atoms in bondseq order.
  **********************************************************/
  dt_reset(bondseq);
  atom2 = dt_xatom(atom, dt_next(bondseq));
  i = dt_integer(atom2, 5, "index");
  a[0] = atoms[i].x; a[1] = atoms[i].y; a[2] = atoms[i].z;

  atom2 = dt_xatom(atom, dt_next(bondseq));
  i = dt_integer(atom2, 5, "index");
  b[0] = atoms[i].x; b[1] = atoms[i].y; b[2] = atoms[i].z;

  atom2 = dt_xatom(atom, dt_next(bondseq));
  i = dt_integer(atom2, 5, "index");
  c[0] = atoms[i].x; c[1] = atoms[i].y; c[2] = atoms[i].z;

  /**********************************************************
  *** If we have an implicit H, use central atom xyz's.
  *** This will work for all 3D cases, but not all conceivable
  *** "2.5D" pictures (i.e., _!_ ).
  **********************************************************/
  if (imp_hcount) i = dt_integer(atom, 5, "index");
  else {
    atom2 = dt_xatom(atom, dt_next(bondseq));
    i = dt_integer(atom2, 5, "index");
  }
  d[0] = atoms[i].x; d[1] = atoms[i].y; d[2] = atoms[i].z;

  return (du_3d_signed_volume(a, b, c, d));
}

/********************************************************************
***  setchiral1 - MDL "parity" -> tetrahedral chirality
*********************************************************************
***
***        3              3            Where:
***        |  parity      |  parity      * = chiral atom
***        *   = 1        *   = 2        4 is behind *
***       / \            / \             1234 is MOL-order
***      2   1          1   2            4 is imp H if present
***
***     4[*@@]321        4[*@]321  <- SMILES equivalent
*********************************************************************
*** Note: parity = 0 -> not chiral; = 3 -> either/unmarked
*********************************************************************
*** Note that explicit hydrogens may be added by this function.
********************************************************************/

static int setchiral1(dt_Handle mol)
{
  dt_Handle atom;             /*** Atom object                    ***/
  dt_Handle atomseq;          /*** Chiral Atom-sequence           ***/
  dt_Handle atomstream;       /*** Atom stream                    ***/
  dt_Handle chibonds;         /*** Chiral-bondseq                 ***/
  dt_Handle bond;             /*** Bond object                    ***/
  dt_Handle bondstream;       /*** Bond stream                    ***/
  dt_Handle hyd;              /*** Chiral hydrogen atom           ***/

  int i, j, k;                /*** Generic indices                ***/
  int imp_hcount = 0;         /*** (Chiral) implicit H-count      ***/
  int rank[4];                /*** (Chiral) rank of atoms         ***/
  
  atomstream = dt_stream(mol, TYP_ATOM);
  atomseq = stream2seq(atomstream);
  dt_dealloc(atomstream);
  while (NULL_OB != (atom = dt_next(atomseq))) {
    i = dt_integer(atom, 5, "index");
    if (1 == atoms[i].parity || 2 == atoms[i].parity) {

      /*** If imp_hcount > 1, cannot be chiral. ***/
      if (1 < (imp_hcount = dt_imp_hcount(atom))) continue;
      if (4 != (dt_count(atom, TYP_BOND) + imp_hcount)) continue;

      /*** Note: rank[] of connected atom is its index. ***/
      for (j = 0; j < 4; rank[j++] = 0)  ; /*** Zero array. ***/
      bondstream = dt_stream(atom, TYP_BOND);
      for (j = 0; NULL_OB != (bond = dt_next(bondstream)); ++j)
        rank[j] = dt_integer(dt_xatom(atom, bond), 5, "index");
      dt_dealloc(bondstream);

      /***************************************************
      *** Create bond-seq in reverse MOL-order.
      *** If present, the implicit hydrogen is first.
      ***************************************************/
      chibonds = dt_alloc_seq();

      /***************************************************
      *** Replace implicit H with explicit.  "Added" H's
      *** have pseudo-index zero in the arrays.
      ***************************************************/
      if (1 == imp_hcount) {
        dt_setimp_hcount(atom, 0);
        hyd = atoms[natoms + ++nnewhs].h = dt_addatom(mol, DX_ATN_H, 0);
	dt_setinteger(hyd, 5, "index", natoms + nnewhs);
        atoms[natoms + nnewhs].x = atoms[i].x;
        atoms[natoms + nnewhs].y = atoms[i].y;
        atoms[natoms + nnewhs].z = atoms[i].z;
        atoms[natoms + nnewhs].parity = 0;
        atoms[natoms + nnewhs].hhh = 0;
        atoms[natoms + nnewhs].visited = 0;
        bond = dt_addbond(atom, hyd, 1);
        dt_append(chibonds, bond);
	dt_setinteger(bond, 5, "index", nbonds + nnewhs);
        bonds[nbonds + nnewhs].h = bond;
        bonds[nbonds + nnewhs].from_atom = atom;
        bonds[nbonds + nnewhs].to_atom = hyd;
        bonds[nbonds + nnewhs].stereo = 0;
      }
      for (k = natoms; k > 0; --k) {
        for (j = 0; j < (4 - imp_hcount); ++j) {
          if (rank[j] == k) {
            dt_append(chibonds, dt_bond(atom, atoms[k].h));
            break;
          }
        }
      }

      /*******************************************************
      *** Assign chirality if 4 bonds and not already set.
      *******************************************************/
      if (4 == (dt_count(chibonds, TYP_BOND)) &&
          DX_CHI_NONE == dt_chival(atom, chibonds)) {
        dt_setchival(atom, chibonds, 
                          atoms[i].parity == 1 ? DX_CHI_THCW : DX_CHI_THCCW);
      }
      dt_dealloc(chibonds);
    }
  }
  dt_dealloc(atomseq);
  return (1);
}

/********************************************************************
***  setchiral2 - MDL "bond_stereo" + 2D -> tetrahedral chirality
***                                or 3D -> tetrahedral chirality
*********************************************************************
***  More tetrahedral chirality, interpreted from bond_stereo
***  Must have four single bonds, at least two up/down to
***  interpret chirality.  Must have at least 2D coords.  
*********************************************************************
***  Note that setchiral2 will set chirality in some cases with
***  one implicit H.  The signed volume is calculated with the
***  implicit H located at the central atom.
*********************************************************************
*** Note that explicit hydrogens may be added by this function.
********************************************************************/

static int setchiral2(dt_Handle mol)
{
  dt_Handle atom, atomseq, atomstream, bond, bondseq, bondstream;
  dt_Handle from_atom, hyd, to_atom;
  float vol;
  int i, j, k, bi, imp_hcount = 0;
  int stereobondcount;
  int wegot_3d;

  wegot_3d = dt_boolean(mol, 5, "got3d");

  atomstream = dt_stream(mol, TYP_ATOM);
  atomseq = stream2seq(atomstream);
  dt_dealloc(atomstream);

  /*********************************************************
  *** Examine every atom for possible chiral center.
  *** Note we must use a seqence instead of a stream since
  *** structural modifications deallocate a stream.
  *********************************************************/
  for (hyd = NULL_OB; NULL_OB != (atom = dt_next(atomseq)); ) {
    stereobondcount = 0;
    bondstream = dt_stream(atom, TYP_BOND);
    bondseq = stream2seq(bondstream);
    dt_dealloc(bondstream);

    i = dt_integer(atom, 5, "index");

    /*********************************************************
    *** Check before proceeding:
    ***  - Not an added explicit H (index 0).
    ***  - At least one bond with stereo (or 3D).
    ***  - No more than one imp H.
    ***  - Four single bonds.
    ***  - If one implicit H, either use 3Ds, or, 2Ds where
    ***    there is one and only one up/down (1/6) bond.
    ***    With 3Ds we put the H at the center,
    ***    and get a good signed_volume.  With 2Ds we place
    ***    the H opposite the up/down atom.
    ***  - Chirality must not already be set.
    *********************************************************/
    for (k = 0; NULL_OB != (bond = dt_next(bondseq)); ) {
      j = bonds[dt_integer(bond, 5, "index")].stereo;
      if (j) ++stereobondcount;
      k += j;
    }
    if ((!i)                                            ||
        (!k && !wegot_3d)                               ||
        (1 < (imp_hcount = dt_imp_hcount(atom)))        ||
        (4 != imp_hcount + dt_count(bondseq, TYP_BOND)) ||
        (imp_hcount && !wegot_3d && stereobondcount != 1)  ||
        (DX_CHI_NONE != dt_chival(atom, bondseq))) {
      dt_dealloc(bondseq);
      continue;
    }
    dt_reset(bondseq);

    /*********************************************************
    *** Replace implicit H with explicit.  
    *** If 3D, locate this H at the possible chiral center.
    *** If no 3Ds, locate H in next block of code.
    *********************************************************/
    if (imp_hcount) {
      dt_setimp_hcount(atom, 0);
      hyd = atoms[natoms + ++nnewhs].h = dt_addatom(mol, DX_ATN_H, 0);
      dt_setinteger(hyd, 5, "index", natoms + nnewhs);
      atoms[natoms+nnewhs].parity = 0;
      atoms[natoms+nnewhs].hhh = 0;
      atoms[natoms+nnewhs].visited = 0;
      bond = bonds[nbonds+nnewhs].h = dt_addbond(atom, hyd, 1);
      dt_setinteger(bond, 5, "index", nbonds + nnewhs);
      bonds[nbonds+nnewhs].from_atom = atom;
      bonds[nbonds+nnewhs].to_atom = hyd;
      bonds[nbonds+nnewhs].stereo = 0;
      if (wegot_3d) {
        atoms[natoms+nnewhs].x = atoms[i].x;
        atoms[natoms+nnewhs].y = atoms[i].y;
        atoms[natoms+nnewhs].z = atoms[i].z;
      }
    }

    /*********************************************************
    *** If no 3D's, assign fake z's to "to-atoms."  Note that
    *** only the "from-atom" derives stereo info for a given
    *** bond.  The from_atom is levelled on the page before
    *** we start in case it was previously dinked.
    **********************************************************
    *** If there is an implicit H, locate it opposite the
    *** one and only one stereo atom (up-down or down-up).
    *********************************************************/
    if (!wegot_3d) {
      atoms[i].z = 0.0;
      while (NULL_OB != (bond = dt_next(bondseq))) {
        bi = dt_integer(bond, 5, "index");

        from_atom = bonds[bi].from_atom;
        to_atom   = bonds[bi].to_atom;
        j = dt_integer(to_atom, 5, "index");

        if (atom == from_atom) { 
          switch (bonds[bi].stereo) {
            case (1): atoms[j].z =  1.0; /*** UP   ***/
              if (imp_hcount) {
                atoms[natoms+nnewhs].x = atoms[j].x;
                atoms[natoms+nnewhs].y = atoms[j].y;
                atoms[natoms+nnewhs].z = -1.0;
              }
              break;
            case (6): atoms[j].z = -1.0; /*** DOWN ***/
              if (imp_hcount) {
                atoms[natoms+nnewhs].x = atoms[j].x;
                atoms[natoms+nnewhs].y = atoms[j].y;
                atoms[natoms+nnewhs].z = 1.0;
              }
              break;
            default: atoms[j].z = 0.0;   break;
          }
        } else {
          atoms[j].z = 0.0;
        }
      }
    }
    vol = signed_volume(atom, bondseq);
    if (fabs(vol) > FLT_EPSILON) {

      /***************************************************
      *** If there's an imp H, we add it to the seq.
      ***************************************************/
      if (imp_hcount) {
        bond = dt_bond(hyd, atom);
        dt_append(bondseq, bond);
      }
      if (0.0 > vol) dt_setchival(atom, bondseq, DX_CHI_THCCW);
      else dt_setchival(atom, bondseq, DX_CHI_THCW);
    }
    dt_dealloc(bondseq);
  }
  dt_dealloc(atomseq);
  return (1);
}

/********************************************************************
***  setcistrans -  CIS/TRANS adjustments from 2D/3D for all
***                 possible sites.
*********************************************************************
***                              We look for double bonded
***      2       4               atoms that are 3-connected.
***       \     /                Replace imp H's with exp H's,
***        0 = 1    ez_atoms[]   and check coordinates for
***       /     \                CIS/TRANS.   
***      3       5               
***                              
*********************************************************************
*** Note that explicit hydrogens may be added by this function.
********************************************************************/

static int setcistrans(dt_Handle mol, int ring_cistrans)
{
  dt_Handle atom;             /*** Atom object                    ***/
  dt_Handle atomstream;       /*** Atom stream                    ***/
  dt_Handle bond;             /*** Bond object                    ***/
  dt_Handle bondseq;          /*** Chiral & CIS/TRANS bond-seq    ***/
  dt_Handle bondstream;       /*** Bond stream                    ***/
  dt_Handle dbond;            /*** CIS/TRANS double bond          ***/
  dt_Handle ez_atoms[6];      /*** CIS/TRANS atoms                ***/
  dt_Handle hyd;              /*** Chiral hydrogen atom           ***/

  float a[3];                 /*** Vector for dihedral calc       ***/
  float b[3];                 /*** Vector for dihedral calc       ***/
  float c[3];                 /*** Vector for dihedral calc       ***/

  int dboval;                 /*** (CIS/TRANS) bond designator    ***/
  int i, j, k, jj;            /*** Generic indices                ***/
  int bi;                     /*** Bond index                     ***/
  int bond_in_ring;           /*** Bond in ring?                  ***/
  int imp_hcount = 0;         /*** (Chiral) implicit H-count      ***/
  int dihedral;               /*** Dihedral angle in degrees      ***/
  int ii[6];                  /*** Indices for CIS/TRANS atoms    ***/

  bondstream = dt_stream(mol, TYP_BOND);
  bondseq = stream2seq(bondstream);
  dt_dealloc(bondstream);
  while (NULL_OB != (dbond = dt_next(bondseq))) {
    bi = dt_integer(dbond, 5, "index");
    if (2 != dt_bondtype(dbond)) continue;

    /*** Stereo == 3 is no-stereo condition. ***/
    if (0 != bonds[bi].stereo) continue;

    /*******************************************************
    *** If CIS/TRANS ring bonds not allowed, temporarily
    *** mod_off to check for cycle membership.
    *******************************************************/
    if (!ring_cistrans) {
      if (!dt_mod_off(mol)) {
        fprintf(stderr, "ERROR (MOL2SMI) setcistrans/dt_mod_off failure...\n");
        return (0);
      }
      bond_in_ring = dt_count(dbond, TYP_CYCLE);
      dt_mod_on(mol);
      if (bond_in_ring) continue;
    }

    atomstream = dt_stream(dbond, TYP_ATOM);
    ez_atoms[0] = dt_next(atomstream);
    ez_atoms[1] = dt_next(atomstream);
    dt_dealloc(atomstream);

    /*** Check for 3-connectedness. ***/
    imp_hcount = dt_imp_hcount(ez_atoms[0]);
    if (3 != (imp_hcount + dt_count(ez_atoms[0], TYP_BOND)))
      continue;
    imp_hcount = dt_imp_hcount(ez_atoms[1]);
    if (3 != (imp_hcount + dt_count(ez_atoms[1], TYP_BOND)))
      continue;

    /********************************************************************
    *** Replace implicit H's with explicit.  Set their adjuncts
    *** (indices) to zero for identification later.
    ********************************************************************/
    for (j = 0; j <= 1; ++j) {
      if (1 == dt_imp_hcount(ez_atoms[j])) {
        atom = ez_atoms[j];
        dt_setimp_hcount(atom, 0);
        hyd = atoms[natoms + ++nnewhs].h = dt_addatom(mol, DX_ATN_H, 0);
        dt_setinteger(hyd, 5, "index", natoms + nnewhs);
	i = dt_integer(atom, 5, "index");
        atoms[natoms + nnewhs].x = atoms[i].x;
        atoms[natoms + nnewhs].y = atoms[i].y;
        atoms[natoms + nnewhs].z = atoms[i].z;
        atoms[natoms + nnewhs].parity = 0;
        atoms[natoms + nnewhs].hhh = 0;
        atoms[natoms + nnewhs].visited = 0;
        bond = bonds[nbonds + nnewhs].h = dt_addbond(atom, hyd, 1);
	dt_setinteger(bond, 5, "index", nbonds + nnewhs);
        bonds[nbonds + nnewhs].from_atom = atom;
        bonds[nbonds + nnewhs].to_atom = hyd;
        bonds[nbonds + nnewhs].stereo = 0;
      }
    }

    /*** Assign atoms #2 and #3 ***/
    bondstream = dt_stream(ez_atoms[0], TYP_BOND);
    for (j = 2; NULL_OB != (bond = dt_next(bondstream)); ++j) {
      if (2 == dt_bondtype(bond)) { --j; continue; }
      ez_atoms[j] = dt_xatom(ez_atoms[0], bond);
    }
    dt_dealloc(bondstream);
    if (j != 4) continue; /*** Check for 2 single bonds. ***/

    /*** Assign atoms #4 and #5 ***/
    bondstream = dt_stream(ez_atoms[1], TYP_BOND);
    for (j = 4; NULL_OB != (bond = dt_next(bondstream)); ++j) {
      if (2 == dt_bondtype(bond)) { --j; continue; }
      ez_atoms[j] = dt_xatom(ez_atoms[1], bond);
    }
    dt_dealloc(bondstream);
    if (j != 6) continue; /*** Check for 2 single bonds. ***/

    /***********************************************************
    ***  If ez_atoms[j=2] was an implicit H, use j=3.
    ***  If ez_atoms[k=4] was an implicit H, use k=5.
    ***********************************************************/
    j = ((dt_integer(ez_atoms[2], 5, "index") <= natoms) ? 2 : 3);
    k = ((dt_integer(ez_atoms[4], 5, "index") <= natoms) ? 4 : 5);

    /*** Atom indices ***/
    for (jj = 0; jj < 6; ++jj) ii[jj] = dt_integer(ez_atoms[jj], 5, "index");

    a[0] = atoms[ii[0]].x - atoms[ii[j]].x;
    a[1] = atoms[ii[0]].y - atoms[ii[j]].y;
    a[2] = atoms[ii[0]].z - atoms[ii[j]].z;

    b[0] = atoms[ii[1]].x - atoms[ii[0]].x;
    b[1] = atoms[ii[1]].y - atoms[ii[0]].y;
    b[2] = atoms[ii[1]].z - atoms[ii[0]].z;

    c[0] = atoms[ii[k]].x - atoms[ii[1]].x;
    c[1] = atoms[ii[k]].y - atoms[ii[1]].y;
    c[2] = atoms[ii[k]].z - atoms[ii[1]].z;

    dihedral = (int) fabs(rad2deg(du_3d_dangle(a, b, c)));
    dboval = (((90 - dihedral) * pow(-1, j + k)) > 0 ?
              DX_CHI_CIS : DX_CHI_TRANS);
    /*** Assign CIS/TRANS stereochemistry. ***/
    dt_setdbo(dt_bond(ez_atoms[0], ez_atoms[1]),
              dt_bond(ez_atoms[0], ez_atoms[j]),
              dt_bond(ez_atoms[1], ez_atoms[k]),
              dboval);
  }
  dt_dealloc(bondseq);
  return (1);
}

/********************************************************************
*** strn2num - converts the first (len) chars of (str) to
***            the type specified by (type) and places the
*** result in the address given by (ptr).
*** Types allowed: INT (1), FLOAT (2), STRING (3)
********************************************************************/

static int strn2num(char *str, int len, int type, void *ptr)
{
  static char *buff = NULL;
  static int buffsize = 0;

  if (!buff) {
    buffsize = len + 1;
    buff = (char *) malloc(buffsize * sizeof(char));
  }
  if (len >= buffsize) {
    buffsize = len + 1;
    buff = (char *) realloc(buff, buffsize * sizeof(char));
  }
  strncpy(buff, str, len);
  buff[len] = '\0';

  switch (type) {
    case (INT):   *((int *) ptr) = atoi(buff);
                  break;
    case (FLOAT): *((float *) ptr) = atof(buff);
                  break;
    case (STRING): strcpy((char *) ptr, buff);
                  break;
    default: fprintf(stderr, "ERROR (strn2num): Unknown type.\n");
             return (0);
  }
  return (1);
}


/*============================================================================
 *   rd_valence() -- return valence for given atom, using .RD rules
 *
 *   This is derived from ACD structures (ignores the published MDL tables).
 */

static int rd_valence(dt_Handle atom, int vbond)
{
   int atno   = dt_number(atom);
   int charge = dt_charge(atom);
   int ncon   = dt_count(atom, TYP_BOND);
   int val;

   switch (dt_number(atom))
     {
     case DX_ATN_WILD:
       val = 0;
       break;
       
       /*** Group 1a ***/
     case DX_ATN_H:  case DX_ATN_Li: case DX_ATN_Na: case DX_ATN_K:  
     case DX_ATN_Rb: case DX_ATN_Cs: case DX_ATN_Fr: 
       val = (charge == 0 ? 1 : 0);
       break;
       
       /*** Group 2a ***/
       
     case DX_ATN_Be: case DX_ATN_Mg: case DX_ATN_Ca: case DX_ATN_Sr: 
     case DX_ATN_Ba: case DX_ATN_Ra: 
       val = 2 - charge;
       break;
       
       /*** Group 3a ***/
       
     case DX_ATN_Al: case DX_ATN_Ga: case DX_ATN_B:
       val = 3 - charge;
       break;
     case DX_ATN_Tl:
       val = 3 - DU_ABS(charge + 2);
       if (vbond > val)
	 val = 5 - DU_ABS(charge + 2);
       break;
     case DX_ATN_In:
       val = 0;
       break;

       /*** Group 4a ***/
       
     case DX_ATN_C:  case DX_ATN_Si: case DX_ATN_Ge: 
       val = 4 - DU_ABS(charge);
       break;
       
     case DX_ATN_Sn:
       val = 3 - DU_ABS(charge + 1);
       if (vbond > val)
	 val = 5 - DU_ABS(charge + 1);
        break;

     case DX_ATN_Pb:
       val = 2 - DU_ABS(charge);
       if (vbond > val)
	 val = 4 - DU_ABS(charge);
       break;
       
       /*** Group 5a ***/

     case DX_ATN_As: 
       val = 4 - DU_ABS(charge - 1);
       if (vbond > val)
	 val = 6 - DU_ABS(charge - 1);
       break;
	   
     case DX_ATN_N:
       val = 3 + charge;
       if (vbond > val)
	 val = 5 + charge;
       break;
       
     case DX_ATN_Bi:
       val = 3 - charge;
       break;
     case DX_ATN_Sb:
       val = 3 - charge;
       if (vbond > val)
	 val = 5 - charge;
       break;
       
     case DX_ATN_P:
       if      (1 == charge) val = 4;
       else if (3 <  vbond ) val = 5;
       else                  val = 3 - charge;
       break;
       
       /*** Group 6a ***/
       
     case DX_ATN_O:  case DX_ATN_Se: case DX_ATN_Te: case DX_ATN_Po: 
       val = 2 + charge;
       break;
       
     case DX_ATN_S:
       if (dt_aromatic(atom)) val = 2;  /* aromatic sulfur has 0 H's */
       else if ( 1 == charge) val = 3;  /* JJD */
       else if (-1 == charge) val = 1;
       else if ( 4 <  vbond ) val = 6;
       else if ( 2 <  vbond ) val = 4;
       else                   val = 2;
       break;
       
       /*** Group 7a ***/

     case DX_ATN_I: case DX_ATN_Cl: case DX_ATN_Br:

       if (charge < 0) val = 0;
       else
	 {
	   val = 1 + charge;
	   if (vbond > val)
	     val = 3 + charge;
	   if (vbond > val)
	     val = 5 + charge;
	 }
       break;

     case DX_ATN_F:
     case DX_ATN_At: 
       val = 1 + charge;
       break;
       
       /*** Group 0 ***/
       
     case DX_ATN_He: case DX_ATN_Ne: case DX_ATN_Ar: case DX_ATN_Kr: 
     case DX_ATN_Xe: case DX_ATN_Rn: 
       val = 0;
       break;
       
       /*** Assume everything else has zero default valence. ***/

     default:
       val = 0;
       break;
     }

   /*** Return valence but not less than zero. ***/
   
   return (0 < val ? val : 0);
}

/*============================================================================
 *  rd_normh() -- return normal H-count for given atom, using .RD rules
 *
 *  Figures normal H-count based on atom's number, charge and connectivity.
 */

static int rd_normh(dt_Handle atom, int fixval, int radcode)
{
   int       val, abond, vbond, atno, hcount;
   dt_Handle bond, bonds;

   /*** Count bond equivalents. ***/

   vbond = 0;
   abond = 0;
   bonds = dt_stream(atom, TYP_BOND);
   while (NULL_OB != (bond = dt_next(bonds)))
     {
       switch (dt_bondtype(bond))
	 {
         case DX_BTY_SINGLE:  vbond += 1;  break;
         case DX_BTY_DOUBLE:  vbond += 2;  break;
         case DX_BTY_TRIPLE:  vbond += 3;  break;
         case DX_BTY_AROMAT:  vbond += 1;  abond += 1;  break;
	 }
     }
   if (0 < abond) vbond++;
   dt_dealloc(bonds);
   
   /*** Get valence if needed. ***/
   
   atno = dt_number(atom);
   val  = (0 > fixval ? rd_valence(atom, vbond) : fixval);

   switch (radcode)
     {
     case 1: case 3:
       val = DU_MAX(0, val - 2); break;
     case 2:
       val = DU_MAX(0, val - 1); break;
     default: break;
     }
   
   /*** Return outstanding valence. ***/
   
   return (DU_MAX(0, val - vbond));
}

/*============================================================================
 *   radical_rings()
 *
 *   Operates in mod-on mode
 */

static int radical_rings(dt_Handle mol)
{
  dt_Handle  atomstream, atom, bonds, bond, cycles, cycle, acharge;
  int i;

  /*** Loop over cycles in molecule. ***/
  
  cycles = dt_stream(mol, TYP_CYCLE);
  while (NULL_OB != (cycle = dt_next(cycles)))
    {
      /*** Loop over atoms in ring. ***/
      
      atomstream = dt_stream(cycle, TYP_ATOM);
      while (NULL_OB != (atom = dt_next(atomstream)))
	{
	  i = dt_integer(atom, 5, "index");
	  if (atoms[i].radcode == 0)
	    break;
	}

      /*** If cycle has all radical atoms, it's a radical cycle. ***/
      if (NULL_OB == atom)
	{
         /*** Set aromatic bonds. ***/
	  bonds = dt_stream(cycle, TYP_BOND);
	  while (NULL_OB != (bond = dt_next(bonds)))
            dt_setbondtype(bond, DX_BTY_AROMAT);
	  dt_dealloc(bonds);

         /*** Set aromatic atoms. ***/
	  acharge = NULL_OB;
	  dt_reset(atomstream);
	  while (NULL_OB != (atom = dt_next(atomstream)))
	    {
	      dt_setaromatic(atom, TRUE);
	      dt_setimp_hcount(atom, rd_normh(atom, -1, 0));
	      if (NULL_OB == acharge && 0 == dt_charge(atom)) acharge = atom;
	    }
	  
	  /*** Now we have cycles with all "radical" atoms: set charge ***/
	  
	  if (acharge)
	    {
            switch (dt_count(cycle, TYP_ATOM))
	      {
	      case 5:  dt_setcharge(acharge, -1);  break;
	      case 7:  dt_setcharge(acharge,  1);  break;
	      }
	    }
	}
      dt_dealloc(atomstream);
    }
  dt_dealloc(cycles);
  return TRUE;
}

