/*#======================================================================
# This source code is public domain, and may be freely distributed,
# modified, and used for any purpose.  IT COMES WITH ABSOLUTELY NO
# WARRANTY OF ANY KIND.
#======================================================================
# _DY_CONTRIB_PGM_:molalign.c: Align 2D layout based on a SMARTS pattern
#======================================================================
# LICENSE MODULES: toolkits:
#                  programs:
#                  servers:
# AUTHORS:  TJ O'Donnell
# DESCRIPTION: 
# Program to modify coords of an input sdf file so that all
# molecules are aligned the same way.  reads a smarts to
# define which atoms are to be aligned.  reads optional rdf
# file for reference coords for alignment.  otherwise aligns
# to first sdf molecule.  optionally aligns to x or y axis.
# reads sdf on stdin and produces modified sdf on stdout.
#
#  Current known limitations: 
#   assumes first product molecule for rdf alignment.
#
#======================================================================*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "dt_smiles.h"
#include "dt_smarts.h"
#include "du_3d.h"
#include "du_mol2smi.h"

int v = 1; /* for printing errors, etc. */
#define MAXATOM 500
#define MOLSIZE 50000
#define MAXSMI 5000

dt_Handle sd2dy ( char *molstr, int *rc, int iso ) {

dt_Handle mol;
FILE *MOLTMP;
char strtmp[200];

MOLTMP = tmpfile();
if (NULL == MOLTMP) {
  fprintf(stderr,"molalign: can't make temporary file, error number %d\n",errno);
  perror("molalign");
  return NULL_OB;
}
fprintf (MOLTMP, "%s", molstr);
rewind (MOLTMP);
fgets(strtmp, sizeof(strtmp), MOLTMP);
fgets(strtmp, sizeof(strtmp), MOLTMP);
fgets(strtmp, sizeof(strtmp), MOLTMP);
mol = du_mol2smi (MOLTMP, rc);
fclose (MOLTMP);
return (mol);
}

int get_rxn (int *nreact, char *rxn_fs) {
char c[200];
int clen;
int nprod;
int line;

/* reads a rxn file.  verifies it has $RXN as first line
   and returns number of reactants and products */
/* leaves file at $MOL block for first reactant */

 line = 0;
 nprod = 0;

 while ( fgets(c, sizeof(c), stdin) ) {
  clen = strlen(c);
  line++;
  if (!strncmp(c, rxn_fs, strlen(rxn_fs)) ) {
   nprod = -1;
   break;
  }

  if ( (line == 1) && strncmp(c, "$RXN\n", 5) ) {
   nprod = 0;
   break;
  }

  else if (line == 5) {
   sscanf(c, "%d %d", nreact, &nprod);
   return (nprod);
  }
 }

return (nprod);
}

int read_molblock (char *molblock, int molsize, int rxn_flag, char *rxn_fs) {
int mollen;
char c[200];
int clen;
int line;

/* reads and returns the next molblock out of an sdf (or rdf) file */

mollen = 0;
line = 0;
while ( fgets(c, sizeof(c), stdin) ) {
 if (rxn_flag && !strncmp(c, rxn_fs, strlen(rxn_fs)) ) {
  mollen = -1;
  break; /* and return */
 }

 line++;
 if (line == 1 && !strncmp(c, "$RXN\n", 5))  {
  fprintf (stderr, "read_sdf: appears to be an rdf file.\nskipping to ", rxn_fs);
  while ( fgets(c, sizeof(c), stdin) ) {
   if ( !strncmp(c, rxn_fs, strlen(rxn_fs)) ) {
    fprintf (stderr, "%s", c);
    fgets(c, sizeof(c), stdin);
    break;
   }
   if (v)  fprintf (stderr, ".");
  }
 }
 clen = strlen(c);
 if (mollen+clen >= molsize) {
  fprintf (stderr, "molblock too big; %d > %d\n", mollen+clen, molsize);
  exit(1);
 } else {
  if (clen > 1) {
   strncpy(&molblock[mollen], c, clen);
   mollen += clen;
  } else if (clen < 0 ) {
   fprintf (stderr, "internal error reading molfile\n");
   exit(1);
  } else {
   strncpy(&molblock[mollen], " \n", 2);
   mollen += 2;
  }
 }

 if (rxn_flag) {
/* end of molblock for rxn file */
  if (!strncmp(c, "M  END\n", 7)) {
   molblock[mollen++] = '\0'; return (mollen);
  }
 }
/* end of molblock for sdf file */
 if (!strncmp(c,  "$$$$\n", 5)) 
 { molblock[mollen++] = '\0'; return (mollen); }

}

return (mollen);
}

void ztran(double *x, double *y, double *z, int ncoord, double dx, double dy) {
int i;

 for (i=0; i<ncoord; i++) {
  x[i] += dx;
  y[i] += dy;
 }
}

void zrot(double *x, double *y, double *z,
  int ncoord, double angle, int iflip) {

double xnew, ynew;
double c,s;
int i;

 c = cos(angle);
 s = sin(angle);
 for (i=0; i<ncoord; i++) {
  xnew =  x[i] * c - y[i] * s;
  ynew =  x[i] * s + y[i] * c;
  if (iflip == 0) {
   x[i] = xnew;
   y[i] = ynew;

/* flip around x axis */
  } else if (iflip == 1) {
   y[i] = -ynew;
   x[i] = xnew;

/* flip around y axis */
  } else if (iflip == 2) {
   x[i] = -xnew;
   y[i] = ynew;
  }
 }

}

double zangle(double *x, double *y, double *z, int *order, int *iflip,
 int atom1, int atom2, int atom3) {

double dy, dx;
int i, ib[3];
double angle;
double xx[3], yy[3], zz[3];

/* find the angle that vector between second and first coord
   makes with the +x axis.  order of coords (what is first, second)
   is contained in order array. */ 

 ib[0] = atom1-1;
 ib[1] = atom2-1;
 ib[2] = atom3-1;

 dx = x[order[ib[0]]] - x[order[ib[1]]];
 dy = y[order[ib[0]]] - y[order[ib[1]]];
 angle = atan2(dy,dx);

 for (i=0; i<3; i++) {
  xx[i] = x[order[ib[i]]] - x[order[ib[1]]];
  yy[i] = y[order[ib[i]]] - y[order[ib[1]]];
  zz[i] = z[order[ib[i]]] - z[order[ib[1]]];

 }
 zrot(xx,yy,zz,3,-angle,0);

 *iflip = 0;
 if (yy[2] < 0.0) *iflip = 1;

/*
 if (v) {
  fprintf (stderr, " -> ");
  for (i=0; i<3; i++) {
   fprintf (stderr, "(%4.2f,%4.2f,%4.2f)", xx[i], yy[i], zz[i]);
  }
  fprintf(stderr, "angle=%4.2f; flip=%d\n", angle*180./PI, *iflip);
 }
*/

 return(angle);
}

int parse_molblock (char *molblock, int *natom, int *nbond,
  double *x, double *y, double *z, int mod_coords) {
int iatom;
char *c;
char *m;
int line;
dt_Handle pathset, paths, p;
char coord_tmp[30];
int i;

/*
 this modifies molblock by changing \n to \0. 
 reads and returns natom, nbond. 
 if mod_coords is <0, coords are ignored.
 if mod_coords is  0, coords in molblock are returned in x,y,z
 if mod_coords is >0 , coords in molblock are replaced with x,y,z
*/

iatom = 0;
line = 0;
m = molblock;
while ( (c = strtok(m, "\n")) != NULL) {
 line++;
/* line 4 has natoms and nbonds on it */
 if (line == 4) {
  sscanf(c, "%d %d", natom, nbond);
 }
 if ( (line > 4) && (iatom < *natom) ) {
  if (mod_coords < 0) {

  } else if (mod_coords > 0) {
/* replace coords */
   sprintf (coord_tmp, "%10.4f%10.4f%10.4f", x[iatom], y[iatom], z[iatom]);
   strncpy (c, coord_tmp, 30);

  } else if (mod_coords == 0) {
/* get (and return in arrays) coords */
   sscanf(c, "%lf %lf %lf", &x[iatom], &y[iatom], &z[iatom]);
  }
  iatom++;
 }
 c[strlen(c)] = '\n'; /* put \n back */
 m = NULL; /* for strtok */
}

return (line);
}

int match_atoms (dt_Handle smarts, dt_Handle mol, int atom_order[]) {

dt_Handle pathset, paths, p;
dt_Handle atoms, a;
dt_Handle oatoms, oa;
dt_String symbol;
dt_Integer symlen;
dt_Boolean iso, addh;
int nmatch;
int i, oi;

 iso = 0;
 addh = 0;

/* match smirks/smarts of product to mol */
 pathset = dt_match(smarts, mol, (dt_Boolean)1);
 if (pathset == NULL_OB) {
  nmatch = -1;
  return(nmatch);
 }
 paths = dt_stream(pathset, TYP_PATH);

 oatoms = dt_origstream(mol, TYP_ATOM);
 while ( p = dt_next(paths) ) {
  nmatch = dt_count(p, TYP_ATOM);
  atoms = dt_stream(p, TYP_ATOM);
  i = 0;
  while ( a = dt_next(atoms) ) {
   symbol = dt_symbol(&symlen, a);
   dt_reset(oatoms);
   oi = 0;
   while ( oa = dt_next(oatoms) ) {
    if (oa == a) break;
    oi++;
   }
   atom_order[i] = oi;
   i++;
  }
  dt_dealloc(atoms);
  dt_dealloc(p);
 }

 dt_dealloc(paths);
 dt_dealloc(pathset);

 return (nmatch);
}

void usage (char *progname) {
 fprintf (stderr, "usage: %s -s SMARTS [-h] [-v] [-fs string] [-1 #] [-2 #] [-3 #] [-a sdf|rdf|x|y] [-o smi|sdf] [-p] [-v]\n", progname); 
 fprintf (stderr, " Reads on stdin:\n");
 fprintf (stderr, " (if -a rdf or -p or -s @rdf) an rdf file with product coords,\n");
 fprintf (stderr, " (if -a rdf or -p or -s @rdf) cat'ed to a unique file separator string\n");
 fprintf (stderr, "  cat'ed to an sdf file with many product molecules.\n");
 fprintf (stderr, " Writes on stdout a sdf file of many product molecules\n");
 fprintf (stderr, "  with coords rotated to conform to target (see -a).\n");
 fprintf (stderr, " REQUIRED:\n");
 fprintf (stderr, "  -s SMARTS: SMARTS defining alignment atoms\n");
 fprintf (stderr, "              or\n");
 fprintf (stderr, "  -s \"@sdf\": SMARTS will be SMILES of first sdf molecule\n");
 fprintf (stderr, "              or\n");
 fprintf (stderr, "  -s \"@rdf\": SMARTS will be SMILES (or variants thereof) of first rdf product\n");
 fprintf (stderr, " Options:\n");
 fprintf (stderr, "  -h: this brief help\n");
 fprintf (stderr, "  -fs string: unique string marking end of rdf file (default END-OF-RDF-FILE)\n");
 fprintf (stderr, "  -1 #: 1st atom in smarts (or in rdf if -s @rdf) to align (default=1)\n");
 fprintf (stderr, "  -2 #: 2nd atom in smarts (or in rdf if -s @rdf) to align (default=2)\n");
 fprintf (stderr, "  -3 #: 3rd atom in smarts (or in rdf if -s @rdf) to align (default=3)\n");
 fprintf (stderr, "  -a: align to [rdf][sdf][x][y]\n");
 fprintf (stderr, "      sdf: align to sdf first molecule coordinates (default)\n");
 fprintf (stderr, "      rdf: align to rdf first product coordinates\n");
 fprintf (stderr, "      x: align to x axis\n");
 fprintf (stderr, "      y: align to y axis\n");
 fprintf (stderr, "  -o: [smi][sdf] output file format\n");
 fprintf (stderr, "      sdf : output sdf (default)\n");
 fprintf (stderr, "      smi: output smiles\n");
 fprintf (stderr, "  -p: include rdf product core in output sdf file (ignored unless -o sdf)\n");
 fprintf (stderr, "  -v: diagnostic verbosity on stderr\n");
}


main (int argc, char **argv) {

dt_String smiles;
dt_Integer smilen;
dt_Integer mollen;
dt_String molstr;
char molblock[MOLSIZE];
dt_Boolean iso;
int imol;
double x[MAXATOM], y[MAXATOM], z[MAXATOM];
double sx[MAXATOM], sy[MAXATOM], sz[MAXATOM];
int natom;
int nbond;
int i,j;
int nlines;
int nreact;
int nprod;
dt_Handle smarts_handle[10];
char smarts[200],*p,*q;
int nmatch;
dt_Handle mol;
int sprod_atom[MAXATOM];
int rc;
int rdf_order[MAXATOM];
int sdf_order[MAXATOM];
int smiout, sdfout;
double rangle, sangle, dangle;
int rflip, sflip, dflip;
int atom1=1, atom2=2, atom3=3;
int rdftoo;
int xalign, yalign;
int sdfalign, rdfalign;
int rdfsmarts;
int sdfsmarts;
char rxn_fs[100];
int badmatch=0;
int nsmarts=0;

/* reads an rxn file to determine product core and it's coordinates */
/* reads an sdf file of products, converts each molblock to smi,
   finds the core in each product and re-orients the coords to
   algin with the coords of the core */
/* outputs the modified sdf file */

 int iarg;

 v = 0;
 smiout = 0;
 sdfout = 1;
 rdftoo = 0;
 rdfalign = 0;
 sdfalign = 1;
 rdfsmarts = 0;
 sdfsmarts = 0;
 yalign = xalign = 0;
 strcpy(rxn_fs, "END-OF-RDF-FILE");
 iarg=1;
 while (iarg < argc) {

/* help */
  if ( !strcmp(argv[iarg], "-h") ) {
   usage(argv[0]);
   exit(0);

/* smarts */
  } else if ( !strcmp(argv[iarg], "-s") ) {
   iarg++;
   if (iarg >= argc) {
    fprintf (stderr, "missing smarts(-s) option\n");
    exit(1);
   }

   /* MAK 2/8/02 dequote (' or "); shell script may pass them in */
   if (NULL == (p = strchr(argv[iarg], '\'')))
     p = strchr(argv[iarg], '"');
   if (NULL == p)
     p = argv[iarg];
   else {
     q = strchr(p+1, *p);
     p++;
   }
   if (NULL == q)
     q = strchr(p, '\0');
   
   strncpy (smarts, p, q-p);
   smarts[sizeof(smarts)-1] = '\0';
   if (strlen(smarts) == sizeof(smarts)-1) {
    fprintf( stderr, "smarts longer than %d\n", sizeof(smarts)-1 );
    exit(1);
   }
   if (!strcmp(smarts, "@rdf")) { rdfsmarts = 1; }
   if (!strcmp(smarts, "@sdf")) { sdfsmarts = 1; }

/* rdf file separator */
  } else if ( !strcmp(argv[iarg], "-fs") ) {
   iarg++;
   if (iarg >= argc) {
    fprintf (stderr, "missing file separator(-fs) option\n");
    exit(1);
   }
   strncpy (rxn_fs, argv[iarg], sizeof(rxn_fs));
   if (strlen(rxn_fs) == 0) {
    fprintf( stderr, "missing file separator (-fs)\n" );
    exit(1);
   }
   rxn_fs[sizeof(rxn_fs)-1] = '\0';
   if (strlen(rxn_fs) == sizeof(rxn_fs)-1) {
    fprintf( stderr, "file separator longer than %d\n", sizeof(rxn_fs)-1 );
    exit(1);
   }

/* alignment atoms */
  } else if ( !strcmp(argv[iarg], "-1") ) {
   iarg++;
   if (iarg >= argc) {
    fprintf (stderr, "missing first atom (-1) option\n");
    exit(1);
   }
   atom1 = atoi(argv[iarg]);
   if (atom1 < 1) {
    fprintf (stderr, "alignment atom \"-1 %s\" must be a positive integer\n",
      argv[iarg]);
    exit(1);
   }
  } else if ( !strcmp(argv[iarg], "-2") ) {
   iarg++;
   if (iarg >= argc) {
    fprintf (stderr, "missing second atom(-2) option\n");
    exit(1);
   }
   atom2 = atoi(argv[iarg]);
   if (atom2 < 1) {
    fprintf (stderr, "alignment atom \"-2 %s\" must be a positive integer\n",
      argv[iarg]);
    exit(1);
   }
  } else if ( !strcmp(argv[iarg], "-3") ) {
   iarg++;
   if (iarg >= argc) {
    fprintf (stderr, "missing third atom(-3) option\n");
    exit(1);
   }
   atom3 = atoi(argv[iarg]);
   if (atom3 < 1) {
    fprintf (stderr, "alignment atom \"-3 %s\" must be a positive integer\n",
      argv[iarg]);
    exit(1);
   }

/* output options */
  } else if ( !strcmp(argv[iarg], "-o") ) {
   iarg++;
   if (iarg >= argc) {
    fprintf (stderr, "missing output(-o) option\n");
    exit(1);
   }
   if ( !strcmp(argv[iarg], "smi") ) {
    smiout = 1; sdfout = 0;
   } else if ( !strcmp(argv[iarg], "sdf") ) {
    smiout = 0; sdfout = 1;
   } else {
    fprintf (stderr, "unknown output(-o) option %s\n", argv[iarg]);
    exit(1);
   }

  } else if ( !strcmp(argv[iarg], "-p") ) {
    rdftoo = 1;

/* alignment options */
  } else if ( !strcmp(argv[iarg], "-a") ) {
   iarg++;
   if (iarg >= argc) {
    fprintf (stderr, "missing alignment(-a) option\n");
    exit(1);
   }
   if ( !strcmp(argv[iarg], "sdf") ) {
    sdfalign = 1; rdfalign = 0;
   } else if ( !strcmp(argv[iarg], "rdf") ) {
    sdfalign = 0; rdfalign = 1;
   } else if ( !strcmp(argv[iarg], "x") ) {
    xalign = 1; yalign = 0;
   } else if ( !strcmp(argv[iarg], "y") ) {
    xalign = 0; yalign = 1;
   } else {
    fprintf (stderr, "unknown align(-a) option %s\n", argv[iarg]);
    exit(1);
   }

/* verbose stderr */
   } else if ( !strcmp(argv[iarg], "-v") ) {
    v = 1;

  } else {
   fprintf (stderr, "unknown option %s\n", argv[iarg]);
   usage(argv[0]);
   exit(0);
  }
  iarg++;
 }

/* check out smarts */
 if (rdfsmarts || sdfsmarts) {

 } else {
  if (strlen(smarts) == 0) {
   fprintf ( stderr, "no SMARTS\n");
   exit(1);

  } else {
/* could be (prodcut)part of a smirks */
   i = strcspn(smarts, ">>");
   if (i < strlen(smarts)) {
    i += 2;
/* guess not */
   } else {
    i = 0;
   }

/* parse smarts */
   if(v) fprintf (stderr, "smarts: <%s>\n", smarts+i);
   smarts_handle[0] = dt_smartin(strlen(smarts)-i, smarts+i);
   if (NULL_OB == smarts_handle[0]) {
    fprintf (stderr, "can't understand smarts %s\n",smarts);
    exit(0);
   } else {
    nsmarts = 1;
   }

  }

 }

/* sanity check for atom1, atom2 and atom3 */
 if (atom1 == atom2 || atom1 == atom3 || atom2 == atom3) {
  fprintf (stderr, "align to atoms %d %d %d\n", atom1, atom2, atom3);
  fprintf (stderr, "alignment atoms -1 %d -2 %d -3 %d may not be equal\n",
    atom1,atom2,atom3);
  exit(1);
 }

/* read (optional) rdf file */
 if (rdfalign || rdftoo || rdfsmarts) {
  if(v) fprintf (stderr, "file separator: <%s>\n", rxn_fs);
  if ( nprod = read_rdf(rxn_fs, &atom1, &atom2, &atom3, smarts_handle, &nsmarts,
   rdf_order, sdfout, rdftoo, xalign, yalign, &rangle, &x, &y, &z, &rflip) < 1) {
   fprintf (stderr, "not an rdf reaction file\n");
   exit(1);
/*
  } else if (nprod == 0) {
   fprintf (stderr, "impossible rdf product smiles/smarts\n");
   exit(1);
*/
  }
 } else {
  if (v) fprintf (stderr, "skip reading rdf file; neither -p nor -a rdf\n");
 }

 if(v) fprintf (stderr, "align to atoms %d %d %d\n", atom1, atom2, atom3);

/* now on to the sdf file of products */
 if (v) fprintf (stderr, "**** reading sdf file of products\n");
 imol = 0;

 while (mollen = read_molblock(molblock, MOLSIZE,0,rxn_fs)) {
/* if (imol > 4) break; */

  iso = 0;
  molstr = molblock;
  mol = sd2dy (molstr, &rc, iso);
  if (NULL_OB == mol) {
   fprintf (stderr, "can't make molecule\n");
   exit(1);
  }
  smiles = dt_cansmiles(&smilen, mol, iso);
  if (smiles == NULL) {
   fprintf (stderr, "can't recognize molecule\n");
   exit(1);
  }
  imol++;
  if (smiout) { printf ("%s\n", smiles); continue; }

/* parse to get coords */
  nlines = parse_molblock(molblock, &natom, &nbond, sx, sy, sz, 0);
  if(v) fprintf (stderr, "molecule %d(%d atoms): %s\n", imol, natom, smiles);
  if (natom < 1 ) {
   fprintf (stderr, "%d atoms in sdf molecule %d\n", natom, imol);
   exit(1);
  }

  if (sdfsmarts && imol == 1) {
   smarts_handle[0] = dt_smartin(strlen(smiles), smiles);
   nsmarts = 1;
   if (v) fprintf(stderr, "using smiles '%s' as smarts\n", smiles);
  }

  for (i=0; i<nsmarts; i++) {
   if ( (nmatch = match_atoms(smarts_handle[i], mol, sdf_order)) > 0 ) break;
   if (v) fprintf (stderr, "sdf product #%d does not match smarts #%d\n",
                            imol, i+1);
  }
  if (nmatch < 0) {
   fprintf (stderr, "sdf product #%d does not match smarts\n", imol);
   badmatch++;

  } else {
   if (nmatch < atom1 || nmatch < atom2 || nmatch < atom3) {
    fprintf (stderr,
     "alignment atoms (%d %d %d) greater than smarts atoms (%d)\n",
     atom1,atom2,atom3,nmatch);
    exit(1);
   }
   if(v) {
    fprintf (stderr, "matched %d:", nmatch);
    for (i=0; i<nmatch; i++) {
     fprintf (stderr, " %d", sdf_order[i]+1);
    }
    fprintf (stderr, "\n");
   }

/* compute rotation angle and rotate coords */
   sangle = zangle(sx,sy,sz,sdf_order,&sflip,atom1,atom2,atom3);
   if (xalign) {
    zrot(sx,sy,sz,natom,-sangle,sflip);
 
   } else if (yalign) {
    if (sflip) sflip=2;
    zrot(sx,sy,sz,natom,(-sangle + PI/2),sflip);
 
   } else if (rdfalign) {
    dflip = abs(rflip - sflip);
    if (dflip) {
     dangle = -rangle - sangle;
    } else {
     dangle =  rangle - sangle;
    }
    zrot(sx,sy,sz,natom,dangle,dflip);

   } else if (sdfalign) {
    if (imol == 1) {
/* save first sdf data for reference */
     rangle = sangle;
     rflip = sflip;
     if (v) {
      fprintf (stderr, "sdf align reference mol %d; angle=%4.2f; flip=%d\n",
        imol, rangle*180./PI, rflip);
     }
    } else {
     dflip = abs(rflip - sflip);
     if (dflip) {
      dangle = -rangle - sangle;
     } else {
      dangle =  rangle - sangle;
     }
     zrot(sx,sy,sz,natom,dangle,dflip);
     if (v) {
      fprintf (stderr, "sdf align %d; angle=%4.2f; flip=%d\n",
        imol, dangle*180./PI, dflip);
     }
    }

   }

  } /* end of good match  if (nmatch < 0) else */

/* now replace coords with rotated ones */
   nlines = parse_molblock(molblock, &natom, &nbond, sx, sy, sz, 1);

  if (nmatch < 3) { printf ("Not aligned! "); }
  printf ("%s", molblock);
 }

 if (badmatch) {
  fprintf (stderr, "%d molecules did not match smarts\n", badmatch);
 }

}

int read_rdf (char *rxn_fs, int *atom1, int *atom2, int *atom3,
  dt_Handle smarts_handle[], int *nsmarts, 
  int rdf_order[], int sdfout, int rdftoo, int xalign, int yalign,
  double *rangle, double *x, double *y, double *z, int *rflip) {

/* also returns rangle, rflip, x[], y[] and z[] */

dt_Handle mol;
dt_String smiles;
dt_Integer smilen;
dt_Integer mollen;
dt_String molstr;
char molblock[MOLSIZE];
dt_Boolean iso;
int imol=0;
double sx[MAXATOM], sy[MAXATOM], sz[MAXATOM];
int natom=0;
int nbond=0;
int i=0;
int j=0;
int nlines=0;
int nreact=0;
int nprod=0;
int nmatch=0;
int rc=0;
int reorder_match_atoms = 0;
int a1,a2,a3;

 if(v) fprintf (stderr, "**** reading rdf file for product coords\n");
 nprod = get_rxn(&nreact, rxn_fs);
 if (nprod < 1) {
  return (nprod);
 } else {
  if(v) fprintf (stderr, "%d reactants; %d products\n", nreact, nprod);

/*  reactants */
  for (i=0; i < nreact; i++) {
/* read the $MOL line separator */
   fgets(molblock, sizeof(molblock), stdin);
   if (strncmp(molblock, "$MOL\n", 5))  {
     fprintf (stderr, "missing $MOL line in rdf file\n");
   }
   mollen = read_molblock(molblock, MOLSIZE,1,rxn_fs);
   if (mollen == -1) break;
   nlines = parse_molblock(molblock, &natom, &nbond,
                           (double *)NULL, (double *)NULL, (double *)NULL, -1);
   if(v) fprintf (stderr, "reactant %d: %d atoms; %d bonds\n",
     i+1, natom, nbond);
  }

/*  products */
  for (i=0; i<nprod; i++) {
/* read the $MOL line separator */
   fgets(molblock, sizeof(molblock), stdin);
   if (strncmp(molblock, "$MOL\n", 5))  {
     fprintf (stderr, "missing $MOL line in rdf file\n");
   }
   mollen = read_molblock(molblock, MOLSIZE,1,rxn_fs);
   if (mollen == -1) break;

/*  only consider first product */
   if (i > 0) {
    if (v) fprintf ( stderr, "product %d ignored\n", i+1);
    continue;
   }
   if(v) fprintf (stderr, "product %d: ", i+1);

   iso = 0;
   molstr = molblock;
   mol = sd2dy (molstr, &rc, iso);
   smiles = dt_cansmiles(&smilen, mol, iso);

/*  get coords */
   nlines = parse_molblock(molblock, &natom, &nbond, x, y, z, 0);
   if(v)
    fprintf (stderr, "%s; %d atoms; %d bonds\n", smiles, natom, nbond);
   if (natom < 1) {
    fprintf (stderr, "not enough atoms in rdf product\n");
    exit(1);
   }
   if (*nsmarts == 0) {
    if(v) fprintf (stderr, "using rdf product smiles as smarts\n");
    *nsmarts = concoct_smarts(smarts_handle, smiles);
    if (*nsmarts == 0) {
     fprintf (stderr, "can't concoct smarts\n");
     exit(0);
    }
    reorder_match_atoms = 1;
   }
   nmatch = match_atoms(smarts_handle[0], mol, rdf_order);
   if (nmatch < 0) {
    fprintf (stderr, "rdf product does not match smarts\n");
    exit(1);
   }
   if (nmatch < *atom1 || nmatch < *atom2 || nmatch < *atom3) {
    fprintf (stderr,
     "alignment atoms (%d %d %d) greater than smarts atoms (%d)\n",
     *atom1,*atom2,*atom3,nmatch);
    exit(1);
   }
   if (reorder_match_atoms) {
    if (v) fprintf (stderr, "matched %d:", nmatch);
    for (j=0; j<nmatch; j++) {
     if (v) fprintf (stderr, " %d", rdf_order[j]+1);
     if (rdf_order[j]+1 == *atom1) a1 = j;
     if (rdf_order[j]+1 == *atom2) a2 = j;
     if (rdf_order[j]+1 == *atom3) a3 = j;
    }
    if (v) {
     fprintf (stderr, "\n");
     fprintf (stderr, "reassigning alignment atoms ");
     fprintf (stderr, "%d,%d,%d ",  *atom1, *atom2, *atom3);
    }
    *atom1 = a1+1; *atom2 = a2+1; *atom3 = a3+1;
    if (v) fprintf (stderr, "%d,%d,%d\n", *atom1, *atom2, *atom3);
   }
   *rangle = zangle(x,y,z,rdf_order,rflip,*atom1,*atom2,*atom3);

   if (sdfout && rdftoo) {
    if (xalign) {
     zrot(x,y,z,natom,-*rangle,*rflip);
    } else if (yalign) {
     if (*rflip) *rflip=2;
     zrot(x,y,z,natom,(-*rangle + PI/2), *rflip);
    }
/*   now replace coords with rotated ones */
    nlines = parse_molblock(molblock, &natom, &nbond, x, y, z, 1);
    printf ("%s", molblock);
    printf ("$$$$\n");
   }
   imol = i;
  }

 } /* end of if (nprod < 1) */


/*  read up to end of rdf file separator */
 if (mollen == -1) {
  fprintf ( stderr, "unexpected end of rdf file\n");
  exit(1);
 } else {
  while ( mollen != -1) {
   mollen = read_molblock(molblock, MOLSIZE,1,rxn_fs);
  }
  if (v) fprintf ( stderr, "end of rdf file at %s\n", rxn_fs);
 }

/* also returns nsmarts, rangle, rflip, x[], y[] and z[] */
return (nprod);
}

int find_atom_positions(int apos[], char *nsmiles) {
/* return apos, a list of integers telling where in the
   smiles string each atom symbol occurs.
   it's length should equal the total number of atoms.
   for bracketed atom symbols, the left/open bracket is pointed to.
*/

int i;
int a=0;
char c;

 for (i=0; i<strlen(nsmiles); i++) {
  c = nsmiles[i];
  if (c == '[') {
   apos[a++] = i;
   while (nsmiles[++i] != ']') {continue;}
  } else if (c == '*') {
   apos[a++] = i;
  } else if (c >= 'a' && c <= 'z') {
   apos[a++] = i;
  } else if (c >= 'A' && c <= 'Z') {
   apos[a++] = i;
   if ( nsmiles[i+1] >= 'a' && nsmiles[i+1] <= 'z') {
    if ( !strncmp(&nsmiles[i],"Br",2) 
      || !strncmp(&nsmiles[i],"Cl",2)
       ) i++; 
    }
  }
 }

return(a);
}

int concoct_smarts(dt_Handle smarts_handle[], char *smiles) {

/* invent likely smarts from smiles */

int nsmarts=0;
char *n;
char nsmiles[MAXSMI];
char msmiles[MAXSMI];
int apos[MAXSMI];
int i,j;
int natom, nbond;
dt_Handle mol;
dt_Handle atoms, atom;
char atom_spec[100];
int alen;

 strcpy(msmiles, smiles);
 strcpy(nsmiles, smiles);

/* first, just use the smiles itself */
 strcpy (nsmiles, smiles);
 smarts_handle[nsmarts] = dt_smartin(strlen(nsmiles), nsmiles);
 if(v) fprintf (stderr, "%d(%d): %s\n",
   nsmarts+1, smarts_handle[nsmarts], nsmiles);
 if (smarts_handle[nsmarts]) nsmarts++;

/* replace terminal atoms with aromatic,aliphatic */
 natom = find_atom_positions(apos, nsmiles);
/*   point_smiles(nsmiles, apos, natom); */
 apos[natom] = strlen(nsmiles);
 mol = dt_smilin(strlen(nsmiles), nsmiles);
 strcpy (msmiles, nsmiles);
 nsmiles[0] = '\0';
/* find terminal atoms */
 i=0;
 atoms = dt_origstream(mol, TYP_ATOM);
 while (atom = dt_next(atoms)) {
  alen = apos[i+1] - apos[i];
  strncpy(atom_spec, msmiles+apos[i], alen);
  atom_spec[alen] = '\0';
/* terminal atom */
  if ( (nbond = dt_count(atom, TYP_BOND)) == 1) {
   if (atom_spec[0] == 'C' && atom_spec[1] != 'l') {
    strcat(nsmiles, "[C,c]");
    if (atom_spec[1]) strcat(nsmiles, &atom_spec[1]);
   } else if (atom_spec[0] == 'N') {
    strcat(nsmiles, "[N,n]");
    if (atom_spec[1]) strcat(nsmiles, &atom_spec[1]);
   } else {
    strcat(nsmiles, atom_spec);
   }
  } else {
   strcat(nsmiles, atom_spec);
  }
  i++;
 }
 dt_dealloc(mol);

 if (strlen(nsmiles)) {
  smarts_handle[nsmarts] = dt_smartin(strlen(nsmiles), nsmiles);
  if(v) fprintf (stderr, "%d(%d): %s\n",
    nsmarts+1, smarts_handle[nsmarts], nsmiles);
  if (smarts_handle[nsmarts]) nsmarts++;
 }

/* replace aromatic nH with just n */
 i=0;
 strcpy (msmiles, nsmiles);
 while ( (n = strstr(msmiles, "[nH]")) ) {
  i++;
  *(n+2) = '\0';
  strcpy(nsmiles, msmiles);
  strcat(nsmiles, n+3);
  strcpy (msmiles, nsmiles);
 }
 if (i) {
  smarts_handle[nsmarts] = dt_smartin(strlen(nsmiles), nsmiles);
  if(v) fprintf (stderr, "%d(%d): %s\n",
    nsmarts+1, smarts_handle[nsmarts], nsmiles);
  if (smarts_handle[nsmarts]) nsmarts++;
 }


return (nsmarts);
}

int point_smiles (char *smiles, int apos[], int natom) {

int matom;
dt_Handle mol;
int i,j;

 mol = dt_smilin(strlen(smiles), smiles);
 matom = dt_count(mol, TYP_ATOM);
 dt_dealloc(mol);

 fprintf (stderr, "%s\n", smiles);
 j=0;
 for (i=0; i<strlen(smiles); i++) {

  if (apos[j] == i) {
   fprintf(stderr, "|");
   j++;

/* terminal atoms */
  } else if (apos[j] == -i) {
   fprintf(stderr, "^");
   j++;

  } else {
   fprintf(stderr, " ");
  }

 }

 fprintf (stderr, "%d %d", natom, matom);
 fprintf (stderr, "\n");
}
